import io
import json
import logging
import os
from typing import Any, Dict, Iterator, List, Literal, Optional

from langchain_core.document_loaders import BaseBlobParser, Blob
from langchain_core.documents import Document
from pypdf import PdfReader, PdfWriter
from pypdf.errors import PdfReadError

from langchain_upstage.utils.response_generator import make_request
from langchain_upstage.utils.value_retriever import get_from_param_or_env

logger = logging.getLogger("pypdf")
logger.setLevel(logging.ERROR)

DOCUMENT_PARSE_BASE_URL = "https://api.upstage.ai/v1/document-digitization"
DEFAULT_NUM_PAGES = 10
DOCUMENT_PARSE_DEFAULT_MODEL = "document-parse"

OutputFormat = Literal["text", "html", "markdown"]
OCR = Literal["auto", "force"]
SplitType = Literal["none", "page", "element"]
Category = Literal[
    "table",
    "figure",
    "chart",
    "heading1",
    "header",
    "footer",
    "caption",
    "paragraph",
    "equation",
    "list",
    "index",
    "footnote",
]


def parse_output(data: dict, output_format: OutputFormat) -> str:
    """
    Parse the output data based on the specified output type.

    Args:
        data: The data to be parsed.
        output_format: The output format to parse the element data into.

    Returns:
        str: The parsed output.

    Raises:
        ValueError: If the output type is invalid.
    """
    content = data["content"]
    if output_format == "text":
        return content["text"]
    elif output_format == "html":
        return content["html"]
    elif output_format == "markdown":
        return content["markdown"]
    else:
        raise ValueError(f"Invalid output type: {output_format}")


class UpstageDocumentParseParser(BaseBlobParser):
    """Upstage Document Parse Parser.

    To use, you should have the environment variable `UPSTAGE_API_KEY`
    set with your API key or pass it as a named parameter to the constructor.

    Example:
        ```python
        from langchain_upstage import UpstageDocumentParseParser

        loader = UpstageDocumentParseParser(split="page", output_format="text")
        ```
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        base_url: str = DOCUMENT_PARSE_BASE_URL,
        model: str = DOCUMENT_PARSE_DEFAULT_MODEL,
        split: SplitType = "none",
        chart_recognition: bool = True,
        ocr: OCR = "auto",
        output_format: OutputFormat = "html",
        coordinates: bool = True,
        base64_encoding: Optional[List[Category]] = None,
    ):
        """
        Initializes an instance of the Upstage class.

        Args:
            api_key (str, optional): The API key for accessing the Upstage API.
                Defaults to None, in which case it will be fetched from the environment
                variable `UPSTAGE_API_KEY`.
            base_url (str, optional): The base URL for accessing the Upstage API.
            model (str): The model to be used for the document parse.
                Defaults to "document-parse".
            split (SplitType, optional): The type of splitting to be applied.
                Defaults to "none" (no splitting).
            ocr (OCRMode, optional): Extract text from images in the document using OCR.
                If the value is "force", OCR is used to extract text from an image. If
                the value is "auto", text is extracted from a PDF. (An error will occur
                if the value is "auto" and the input is NOT in PDF format)
            output_format (OutputFormat, optional): Format of the inference results.
            coordinates (bool, optional): Whether to include the coordinates of the
                OCR in the output.
            base64_encoding (List[Category], optional): The category of the elements to
                be encoded in base64.
        """
        self.api_key = get_from_param_or_env(
            "UPSTAGE_API_KEY",
            api_key,
            "UPSTAGE_API_KEY",
            os.environ.get("UPSTAGE_API_KEY"),
        )
        self.base_url = base_url
        self.model = model
        self.split = split
        self.chart_recognition = chart_recognition
        self.ocr = ocr
        self.output_format = output_format
        self.coordinates = coordinates
        self.base64_encoding = base64_encoding if base64_encoding is not None else []

    def _get_headers(self) -> Dict[str, str]:
        """Get headers for API requests with `x-upstage-client` always set to
        `'langchain'`.

        Returns:
            Dict containing Authorization and `x-upstage-client` headers.
        """
        return {
            "Authorization": f"Bearer {self.api_key}",
            "x-upstage-client": "langchain",
        }

    def _get_response(self, files: Dict) -> List:
        """Sends a POST request to the API endpoint with the provided files and returns
        the response.

        Args:
            files: The files to be sent in the request.

        Returns:
            dict: The JSON response from the API.

        Raises:
            ValueError: If there is an error in the API call.
        """
        response = make_request(
            "POST",
            self.base_url,
            self.api_key,
            headers=self._get_headers(),
            files=files,
            data={
                "model": self.model,
                "chart_recognition": self.chart_recognition,
                "ocr": self.ocr,
                "output_formats": f"['{self.output_format}']",
                "coordinates": self.coordinates,
                "base64_encoding": json.dumps(self.base64_encoding),
            },
        )

        return response.get("elements", [])

    def _split_and_request(
        self,
        full_docs: PdfReader,
        start_page: int,
        num_pages: int = DEFAULT_NUM_PAGES,
    ) -> List:
        """Splits the full pdf document into partial pages and sends a request to the
        server.

        Args:
            full_docs (PdfReader): The full document to be split and requested.
            start_page (int): The starting page number for splitting the document.
            num_pages (int, optional): The number of pages to split the document into.

        Returns:
            response: The response from the server.
        """
        merger = PdfWriter()
        merger.append(
            full_docs,
            pages=(start_page, min(start_page + num_pages, full_docs.get_num_pages())),
        )

        with io.BytesIO() as buffer:
            merger.write(buffer)
            buffer.seek(0)
            response = self._get_response({"document": buffer})

        return response

    def _element_document(self, elements: Dict, start_page: int = 0) -> Document:
        """Converts an elements into a `Document` object.

        Args:
            elements (Dict) : The elements to convert.
            start_page (int): The starting page number for splitting the document.

                This number starts from zero.

        Returns:
            A list containing a single `Document` object.

        """
        metadata = {
            "id": elements["id"],
            "page": elements["page"] + start_page,
            "category": elements["category"],
        }

        if self.coordinates and elements.get("coordinates"):
            metadata["coordinates"] = elements.get("coordinates")
        if self.base64_encoding and elements.get("base64_encoding"):
            metadata["base64_encoding"] = elements.get("base64_encoding")

        return Document(
            page_content=(parse_output(elements, self.output_format)),
            metadata=metadata,
        )

    def _page_document(self, elements: List, start_page: int = 0) -> List[Document]:
        """Combines elements with the same page number into a single Document object.

        Args:
            elements: A list of elements containing page numbers.
            start_page: The starting page number for splitting the document.

                This number starts from zero.

        Returns:
            `Document` objects, each representing a page with its content and
                metadata.
        """
        _docs = []
        pages = sorted(set(map(lambda x: x["page"], elements)))

        page_group = [
            [element for element in elements if element["page"] == x] for x in pages
        ]

        for group in page_group:
            page_content = " ".join(
                [parse_output(element, self.output_format) for element in group]
            )

            metadata = {
                "page": group[0]["page"] + start_page,
            }

            if self.base64_encoding:
                base64_encodings = [
                    element.get("base64_encoding")
                    for element in group
                    if element.get("base64_encoding") is not None
                ]
                metadata["base64_encodings"] = base64_encodings

            if self.coordinates:
                coordinates = [
                    element.get("coordinates")
                    for element in group
                    if element.get("coordinates") is not None
                ]
                metadata["coordinates"] = coordinates

            _docs.append(
                Document(
                    page_content=page_content,
                    metadata=metadata,
                )
            )

        return _docs

    def lazy_parse(self, blob: Blob, is_batch: bool = False) -> Iterator[Document]:
        """Lazily parses a document and yields Document objects based on the specified
        split type.

        Args:
            blob (Blob): The input document blob to parse.
            is_batch (bool, optional): Whether to parse the document in batches.
                Defaults to `False` (single page parsing)

        Yields:
            Document: The parsed document object.

        Raises:
            ValueError: If an invalid split type is provided.
        """

        if is_batch:
            num_pages = DEFAULT_NUM_PAGES
        else:
            num_pages = 1

        try:
            full_docs = PdfReader(str(blob.path))
            number_of_pages = full_docs.get_num_pages()
            is_pdf = True
        except PdfReadError:
            number_of_pages = 1
            is_pdf = False
        except Exception as e:
            raise ValueError(f"Failed to read PDF file: {e}")

        if self.split == "none":
            result = ""
            base64_encodings = []
            coordinates = []

            if is_pdf:
                start_page = 0
                num_pages = DEFAULT_NUM_PAGES
                for _ in range(number_of_pages):
                    if start_page >= number_of_pages:
                        break

                    elements = self._split_and_request(full_docs, start_page, num_pages)
                    for element in elements:
                        result += parse_output(element, self.output_format)
                        if self.base64_encoding:
                            base64_encoding = element.get("base64_encoding")
                            if base64_encoding is not None:
                                base64_encodings.append(base64_encoding)
                        if self.coordinates:
                            coordinate = element.get("coordinates")
                            if coordinate is not None:
                                coordinates.append(coordinate)

                    start_page += num_pages

            else:
                if not blob.path:
                    raise ValueError("Blob path is required for non-PDF files.")

                with open(blob.path, "rb") as f:
                    elements = self._get_response({"document": f})

                for element in elements:
                    result += parse_output(element, self.output_format)

                    if (
                        self.base64_encoding
                        and element.get("base64_encoding") is not None
                    ):
                        base64_encoding = element.get("base64_encoding")
                        if base64_encoding is not None:
                            base64_encodings.append(base64_encoding)
                    if self.coordinates and element.get("coordinates") is not None:
                        coordinate = element.get("coordinates")
                        if coordinate is not None:
                            coordinates.append(coordinate)
            metadata: Dict[str, Any] = {
                "total_pages": number_of_pages,
            }
            if self.coordinates:
                metadata["coordinates"] = coordinates
            if self.base64_encoding:
                metadata["base64_encodings"] = base64_encodings

            yield Document(
                page_content=result,
                metadata=metadata,
            )

        elif self.split == "element":
            if is_pdf:
                start_page = 0
                for _ in range(number_of_pages):
                    if start_page >= number_of_pages:
                        break

                    elements = self._split_and_request(full_docs, start_page, num_pages)
                    for element in elements:
                        yield self._element_document(element, start_page)

                    start_page += num_pages

            else:
                if not blob.path:
                    raise ValueError("Blob path is required for non-PDF files.")
                with open(blob.path, "rb") as f:
                    elements = self._get_response({"document": f})

                for element in elements:
                    yield self._element_document(element)

        elif self.split == "page":
            if is_pdf:
                start_page = 0
                for _ in range(number_of_pages):
                    if start_page >= number_of_pages:
                        break

                    elements = self._split_and_request(full_docs, start_page, num_pages)
                    yield from self._page_document(elements, start_page)

                    start_page += num_pages
            else:
                if not blob.path:
                    raise ValueError("Blob path is required for non-PDF files.")
                with open(blob.path, "rb") as f:
                    elements = self._get_response({"document": f})

                yield from self._page_document(elements)

        else:
            raise ValueError(f"Invalid split type: {self.split}")



import os
from typing import Optional


def get_from_param_or_env(
    key: str,
    param: Optional[str] = None,
    env_key: Optional[str] = None,
    default: Optional[str] = None,
) -> str:
    """Get a value from a param or an environment variable."""
    if param is not None:
        return param
    elif env_key and env_key in os.environ and os.environ[env_key]:
        return os.environ[env_key]
    elif default is not None:
        return default
    else:
        raise ValueError(
            f"Did not find {key}, please add an environment variable"
            f" `{env_key}` which contains it, or pass"
            f"  `{key}` as a named parameter."
        )



from langchain_upstage.chat_models import ChatUpstage
from langchain_upstage.document_parse import UpstageDocumentParseLoader
from langchain_upstage.document_parse_parsers import UpstageDocumentParseParser
from langchain_upstage.embeddings import UpstageEmbeddings
from langchain_upstage.tools.prebuilt_information_extraction import (
    UpstagePrebuiltInformationExtraction,
)
from langchain_upstage.tools.universal_information_extraction import (
    UpstageUniversalInformationExtraction,
)

__all__ = [
    "ChatUpstage",
    "UpstageEmbeddings",
    "UpstageDocumentParseLoader",
    "UpstageDocumentParseParser",
    "UpstageUniversalInformationExtraction",
    "UpstagePrebuiltInformationExtraction",
]



import json as json_type
from typing import Any, Optional

import requests


def _default_headers(api_key: str) -> dict:
    return {
        "Authorization": f"Bearer {api_key}",
        "x-upstage-client": "langchain",
    }


def make_request(
    method: str,
    url: str,
    api_key: str,
    headers: Optional[dict[str, str]] = None,
    files: Optional[dict[str, Any]] = None,
    json: Optional[dict[str, Any]] = None,
    data: Optional[dict[str, Any]] = None,
    params: Optional[dict[str, Any]] = None,
) -> dict:
    try:
        response = requests.request(
            method=method,
            url=url,
            headers=(headers or {}) | _default_headers(api_key),
            files=files,
            json=json,
            data=data,
            params=params,
        )

        response.raise_for_status()
        return response.json()

    except requests.HTTPError as e:
        raise ValueError(f"HTTP error: {e.response.text}")
    except requests.RequestException as e:
        # Handle any request-related exceptions
        raise ValueError(f"Failed to send request: {e}")
    except json_type.JSONDecodeError as e:
        # Handle JSON decode errors
        raise ValueError(f"Failed to decode JSON response: {e}")
    except Exception as e:
        # Handle any other exceptions
        raise ValueError(f"An error occurred: {e}")



from __future__ import annotations

import os
from typing import Any, Literal, Optional, Type

from langchain_core.callbacks import (
    AsyncCallbackManagerForToolRun,
    CallbackManagerForToolRun,
)
from langchain_core.tools import BaseTool
from langchain_core.utils import convert_to_secret_str
from pydantic import BaseModel, Field, SecretStr

try:
    import httpx
except ImportError:
    raise ImportError(
        "httpx is required for Prebuilt Information Extraction. "
        "Please install it with: pip install httpx"
    )

from langchain_upstage.utils.constants import MEGABYTE

PREBUILT_EXTRACT_BASE_URL = "https://api.upstage.ai/v1/information-extraction"
MODELS = Literal[
    "receipt-extraction",
    "air-waybill-extraction",
    "bill-of-lading-and-shipping-request-extraction",
    "commercial-invoice-and-packing-list-extraction",
    "kr-export-declaration-certificate-extraction",
]
SUPPORTED_EXTENSIONS = [
    "jpeg",
    "png",
    "bmp",
    "pdf",
    "tiff",
    "heic",
    "docx",
    "pptx",
    "xlsx",
    "hwp",
    "hwpx",
]
MAX_FILE_SIZE_MB = 50
MAX_FILE_SIZE_BYTES = MAX_FILE_SIZE_MB * MEGABYTE


class UpstagePrebuiltInformationExtractionInput(BaseModel):
    """Input for the Prebuilt Information Extraction tool."""

    filename: str = Field(
        description="Path to the document file to extract information from. "
        "Supported formats: jpeg, png, bmp, pdf, tiff, heic, docx, pptx, "
        "xlsx, hwp, hwpx. Maximum file size: 50MB."
    )


def _validate_file(file_path: str) -> None:
    """Validate file exists, format, and size.

    Args:
        file_path: Path to the file to validate.

    Raises:
        FileNotFoundError: If the file does not exist.
        ValueError: If the file format is unsupported or file size exceeds limit.
    """
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"File not found: {file_path}")

    ext = file_path.lower().split(".")[-1]
    if ext not in SUPPORTED_EXTENSIONS:
        supported = ", ".join([f".{e}" for e in SUPPORTED_EXTENSIONS])
        raise ValueError(
            f"Unsupported file format: .{ext}. Supported formats: {supported}"
        )

    file_size = os.path.getsize(file_path)
    if file_size > MAX_FILE_SIZE_BYTES:
        max_size_mb = MAX_FILE_SIZE_BYTES / MEGABYTE
        raise ValueError(
            f"File too large: {file_size / MEGABYTE:.2f}MB. "
            f"Maximum size: {max_size_mb}MB"
        )


class UpstagePrebuiltInformationExtraction(BaseTool):
    """Tool for extracting structured information from documents using prebuilt models.

    This tool uses Upstage's prebuilt information extraction models for specific
    document types such as receipts, air waybills, bills of lading, etc.

    Available models:
        - receipt-extraction: Extract information from receipts
        - air-waybill-extraction: Extract information from air waybills
        - bill-of-lading-and-shipping-request-extraction: Extract from bills of lading
        - commercial-invoice-and-packing-list-extraction: Extract from invoices and
            packing lists
        - kr-export-declaration-certificate-extraction: Extract from Korean export
            certificates

    To use, you should have the environment variable `UPSTAGE_API_KEY`
    set with your API key or pass it as a named parameter to the constructor.

    Args:
        model: The prebuilt model to use for extraction. Must be one of the
            available models.
        api_key: Optional API key. If not provided, will use UPSTAGE_API_KEY
            environment variable.

    Example:
        .. code-block:: python

            from langchain_upstage import UpstagePrebuiltInformationExtraction

            tool = UpstagePrebuiltInformationExtraction(
                model='receipt-extraction'
            )
            result = tool.invoke({"filename": "/path/to/receipt.pdf"})
    """

    name: str = "prebuilt_information_extraction"
    description: str = (
        "Extract structured information from documents using prebuilt extraction "
        "models. Available models: receipt-extraction, air-waybill-extraction, "
        "bill-of-lading-and-shipping-request-extraction, "
        "commercial-invoice-and-packing-list-extraction, "
        "kr-export-declaration-certificate-extraction. "
        "Supports specific document types: receipts, air waybills, "
        "bills of lading, commercial invoices, packing lists, and Korean export "
        "declaration certificates. "
        "Use this tool when you need to extract information from a specific type of "
        "document that matches one of the prebuilt models."
    )

    model: MODELS
    upstage_api_key: Optional[SecretStr] = Field(default=None, alias="api_key")

    args_schema: Type[BaseModel] = UpstagePrebuiltInformationExtractionInput

    def __init__(
        self,
        model: MODELS,
        api_key: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        upstage_api_key = kwargs.get("upstage_api_key", None)
        if not upstage_api_key:
            # api_key is an explicit parameter, not in kwargs
            upstage_api_key = api_key
        if not upstage_api_key:
            upstage_api_key = kwargs.get("api_key", None)
        if not upstage_api_key:
            upstage_api_key = SecretStr(os.getenv("UPSTAGE_API_KEY", ""))
        upstage_api_key = convert_to_secret_str(upstage_api_key)

        if (
            not upstage_api_key
            or not upstage_api_key.get_secret_value()
            or upstage_api_key.get_secret_value() == ""
        ):
            raise ValueError("UPSTAGE_API_KEY must be set or passed")

        super().__init__(
            model=model,
            upstage_api_key=upstage_api_key,
            **kwargs,
        )
        # Ensure upstage_api_key is set (BaseTool might not preserve it)
        self.upstage_api_key = upstage_api_key

    def _run(
        self,
        filename: str,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> dict:
        """Extract information from a document using the prebuilt extraction model.

        Args:
            filename: Path to the document file to extract information from.
            run_manager: Optional callback manager for tool execution.

        Returns:
            dict: Extracted information as a dictionary with fields, documentType, etc.

        Raises:
            FileNotFoundError: If the file does not exist.
            ValueError: If file format is unsupported or file size exceeds limit.
            Exception: If API call fails or response parsing fails.
        """
        try:
            # Validate file
            _validate_file(filename)

            # Get API key
            api_key = (
                self.upstage_api_key.get_secret_value()
                if self.upstage_api_key
                else None
            )
            if not api_key:
                raise ValueError("UPSTAGE_API_KEY must be set or passed")

            # Prepare multipart/form-data request
            with open(filename, "rb") as f:
                files = {
                    "document": (
                        os.path.basename(filename),
                        f,
                        "application/octet-stream",
                    )
                }
                data = {"model": self.model}

                # Make HTTP request
                headers = {
                    "Authorization": f"Bearer {api_key}",
                }

                with httpx.Client(timeout=180.0) as client:  # 3 minute timeout
                    response = client.post(
                        PREBUILT_EXTRACT_BASE_URL,
                        headers=headers,
                        files=files,
                        data=data,
                    )
                    response.raise_for_status()
                    return response.json()

        except (FileNotFoundError, ValueError) as e:
            # Re-raise validation errors as-is
            if run_manager:
                run_manager.on_tool_error(e)
            raise
        except httpx.HTTPStatusError as e:
            # Handle HTTP errors
            error_msg = (
                f"Failed to extract information from document: "
                f"{e.response.status_code} - {e.response.text}"
            )
            if run_manager:
                run_manager.on_tool_error(Exception(error_msg))
            raise Exception(error_msg) from e
        except Exception as e:
            # Wrap other exceptions with context
            error_msg = f"Failed to extract information from document: {str(e)}"
            if run_manager:
                run_manager.on_tool_error(Exception(error_msg))
            raise Exception(error_msg) from e

    async def _arun(
        self,
        filename: str,
        run_manager: Optional[AsyncCallbackManagerForToolRun] = None,
    ) -> dict:
        """Extract information from a document asynchronously using the prebuilt
        extraction model.

        Args:
            filename: Path to the document file to extract information from.
            run_manager: Optional async callback manager for tool execution.

        Returns:
            dict: Extracted information as a dictionary with fields, documentType, etc.

        Raises:
            FileNotFoundError: If the file does not exist.
            ValueError: If file format is unsupported or file size exceeds limit.
            Exception: If API call fails or response parsing fails.
        """
        try:
            # Validate file
            _validate_file(filename)

            # Get API key
            api_key = (
                self.upstage_api_key.get_secret_value()
                if self.upstage_api_key
                else None
            )
            if not api_key:
                raise ValueError("UPSTAGE_API_KEY must be set or passed")

            # Prepare multipart/form-data request
            with open(filename, "rb") as f:
                files = {
                    "document": (
                        os.path.basename(filename),
                        f,
                        "application/octet-stream",
                    )
                }
                data = {"model": self.model}

                # Make HTTP request
                headers = {
                    "Authorization": f"Bearer {api_key}",
                }

                async with httpx.AsyncClient(
                    timeout=180.0
                ) as client:  # 3 minute timeout
                    response = await client.post(
                        PREBUILT_EXTRACT_BASE_URL,
                        headers=headers,
                        files=files,
                        data=data,
                    )
                    response.raise_for_status()
                    return response.json()

        except (FileNotFoundError, ValueError) as e:
            # Re-raise validation errors as-is
            if run_manager:
                await run_manager.on_tool_error(e)
            raise
        except httpx.HTTPStatusError as e:
            # Handle HTTP errors
            error_msg = (
                f"Failed to extract information from document: "
                f"{e.response.status_code} - {e.response.text}"
            )
            if run_manager:
                await run_manager.on_tool_error(Exception(error_msg))
            raise Exception(error_msg) from e
        except Exception as e:
            # Wrap other exceptions with context
            error_msg = f"Failed to extract information from document: {str(e)}"
            if run_manager:
                await run_manager.on_tool_error(Exception(error_msg))
            raise Exception(error_msg) from e



from __future__ import annotations

import logging
import warnings
from typing import (
    Any,
    Dict,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

import openai
from langchain_core.embeddings import Embeddings
from langchain_core.utils import from_env, get_pydantic_field_names, secret_from_env
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    SecretStr,
    model_validator,
)
from typing_extensions import Self

logger = logging.getLogger(__name__)

DEFAULT_EMBED_BATCH_SIZE = 10
MAX_EMBED_BATCH_SIZE = 100

# Constants for header management
UPSTAGE_CLIENT_HEADER = "x-upstage-client"
UPSTAGE_CLIENT_VALUE = "langchain"
DEFAULT_HEADERS = {UPSTAGE_CLIENT_HEADER: UPSTAGE_CLIENT_VALUE}


class UpstageEmbeddings(BaseModel, Embeddings):
    """UpstageEmbeddings embedding model.

    To use, set the environment variable `UPSTAGE_API_KEY` with your API key or
    pass it as a named parameter to the constructor.

    Example:
        ```python
        from langchain_upstage import UpstageEmbeddings

        model = UpstageEmbeddings(model='solar-embedding-1-large')
        ```
    """

    client: Any = Field(default=None, exclude=True)  #: :meta private:

    async_client: Any = Field(default=None, exclude=True)  #: :meta private:

    model: str = Field(...)
    """Embeddings model name to use. Do not add suffixes like `-query` and `-passage`.
    
    Instead, use `'solar-embedding-1-large'` for example.
    """

    dimensions: Optional[int] = None
    """The number of dimensions the resulting output embeddings should have.
    
    Not yet supported. 
    """

    upstage_api_key: SecretStr = Field(
        default_factory=secret_from_env(
            "UPSTAGE_API_KEY",
            error_message=(
                "You must specify an api key. "
                "You can pass it an argument as `api_key=...` or "
                "set the environment variable `UPSTAGE_API_KEY`."
            ),
        ),
        alias="api_key",
    )
    """Automatically inferred from env are `UPSTAGE_API_KEY` if not provided."""

    upstage_api_base: Optional[str] = Field(
        default_factory=from_env(
            "UPSTAGE_API_BASE", default="https://api.upstage.ai/v1/solar"
        ),
        alias="base_url",
    )
    """Endpoint URL to use."""

    embedding_ctx_length: int = 4096
    """The maximum number of tokens to embed at once.
    
    Not yet supported.
    """

    embed_batch_size: int = DEFAULT_EMBED_BATCH_SIZE

    allowed_special: Union[Literal["all"], Set[str]] = set()
    """Not yet supported."""

    disallowed_special: Union[Literal["all"], Set[str], Sequence[str]] = "all"
    """Not yet supported."""

    chunk_size: int = 1000
    """Maximum number of texts to embed in each batch.
    
    Not yet supported.
    """

    max_retries: int = 2
    """Maximum number of retries to make when generating."""

    request_timeout: Optional[Union[float, Tuple[float, float], Any]] = Field(
        default=None, alias="timeout"
    )
    """Timeout for requests to Upstage embedding API. Can be float, httpx.Timeout or
        None."""

    show_progress_bar: bool = False
    """Whether to show a progress bar when embedding.
    
    Not yet supported.
    """
    model_kwargs: Dict[str, Any] = Field(default_factory=dict)

    """Holds any model parameters valid for `create` call not explicitly specified."""

    skip_empty: bool = False
    """Whether to skip empty strings when embedding or raise an error.
    Defaults to not skipping.
    
    Not yet supported."""

    default_headers: Union[Mapping[str, str], None] = DEFAULT_HEADERS
    """add trace header."""

    default_query: Union[Mapping[str, object], None] = None
    # Configure a custom httpx client. See the
    # [httpx documentation](https://www.python-httpx.org/api/#client) for more details.

    http_client: Union[Any, None] = None
    """Optional `httpx.Client`.
    
    Only used for sync invocations.
    
    Must specify  `http_async_client` as well if you'd like a custom client for async
    invocations.
    """

    http_async_client: Union[Any, None] = None
    """Optional httpx.AsyncClient.
    
    Only used for async invocations.
    
    Must specify `http_client` as well if you'd like a custom client for sync
    invocations.
    """

    model_config = ConfigDict(
        extra="forbid",
        populate_by_name=True,
        protected_namespaces=(),
    )

    @model_validator(mode="before")
    @classmethod
    def build_extra(cls, values: Dict[str, Any]) -> Any:
        """Build extra kwargs from additional params that were passed in."""
        all_required_field_names = get_pydantic_field_names(cls)
        extra = values.get("model_kwargs", {})
        for field_name in list(values):
            if field_name in extra:
                raise ValueError(f"Found {field_name} supplied twice.")
            if field_name not in all_required_field_names:
                warnings.warn(
                    f"""WARNING! {field_name} is not default parameter.
                    {field_name} was transferred to model_kwargs.
                    Please confirm that {field_name} is what you intended."""
                )
                extra[field_name] = values.pop(field_name)

        invalid_model_kwargs = all_required_field_names.intersection(extra.keys())
        if invalid_model_kwargs:
            raise ValueError(
                f"Parameters {invalid_model_kwargs} should be specified explicitly. "
                f"Instead they were passed in as part of `model_kwargs` parameter."
            )

        values["model_kwargs"] = extra
        return values

    @model_validator(mode="after")
    def validate_environment(self) -> Self:
        """Validate that api key and python package exists in environment."""

        # Ensure x-upstage-client header is always set to "langchain"
        if self.default_headers is None:
            self.default_headers = DEFAULT_HEADERS
        else:
            # Create a copy to avoid modifying the original
            headers = dict(self.default_headers)
            headers[UPSTAGE_CLIENT_HEADER] = UPSTAGE_CLIENT_VALUE
            self.default_headers = headers

        client_params: dict = {
            "api_key": (
                self.upstage_api_key.get_secret_value()
                if self.upstage_api_key
                else None
            ),
            "base_url": self.upstage_api_base,
            "timeout": self.request_timeout,
            "max_retries": self.max_retries,
            "default_headers": self.default_headers,
            "default_query": self.default_query,
        }
        if not (self.client or None):
            sync_specific: dict = {"http_client": self.http_client}
            self.client = openai.OpenAI(**client_params, **sync_specific).embeddings
        if not (self.async_client or None):
            async_specific: dict = {"http_client": self.http_async_client}
            self.async_client = openai.AsyncOpenAI(
                **client_params, **async_specific
            ).embeddings
        return self

    @property
    def _invocation_params(self) -> Dict[str, Any]:
        self.model = self.model.replace("-query", "").replace("-passage", "")

        params: Dict = {"model": self.model, **self.model_kwargs}
        if self.dimensions is not None:
            params["dimensions"] = self.dimensions
        return params

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Embed a list of document texts using passage model.

        Args:
            texts: The list of texts to embed.

        Returns:
            List of embeddings, one for each text.
        """
        assert self.embed_batch_size <= MAX_EMBED_BATCH_SIZE, (
            f"The embed_batch_size should not be larger than {MAX_EMBED_BATCH_SIZE}."
        )
        if not texts:
            return []
        params = self._invocation_params
        params["model"] = params["model"] + "-passage"
        embeddings = []

        batch_size = min(self.embed_batch_size, len(texts))
        for i in range(0, len(texts), batch_size):
            batch = texts[i : i + batch_size]
            data = self.client.create(input=batch, **params).data
            embeddings.extend([r.embedding for r in data])

        return embeddings

    def embed_query(self, text: str) -> List[float]:
        """Embed query text using query model.

        Args:
            text: The text to embed.

        Returns:
            Embedding for the text.
        """
        params = self._invocation_params
        params["model"] = params["model"] + "-query"

        response = self.client.create(input=text, **params)

        if not isinstance(response, dict):
            response = response.model_dump()
        return response["data"][0]["embedding"]

    async def aembed_documents(self, texts: List[str]) -> List[List[float]]:
        """Embed a list of document texts using passage model asynchronously.

        Args:
            texts: The list of texts to embed.

        Returns:
            List of embeddings, one for each text.
        """
        assert self.embed_batch_size <= MAX_EMBED_BATCH_SIZE, (
            f"The embed_batch_size should not be larger than {MAX_EMBED_BATCH_SIZE}."
        )
        if not texts:
            return []
        params = self._invocation_params
        params["model"] = params["model"] + "-passage"
        embeddings = []

        batch_size = min(self.embed_batch_size, len(texts))
        for i in range(0, len(texts), batch_size):
            batch = texts[i : i + batch_size]
            response = await self.async_client.create(input=batch, **params)
            embeddings.extend([r.embedding for r in response.data])
        return embeddings

    async def aembed_query(self, text: str) -> List[float]:
        """Asynchronous Embed query text using query model.

        Args:
            text: The text to embed.

        Returns:
            Embedding for the text.
        """
        params = self._invocation_params
        params["model"] = params["model"] + "-query"

        response = await self.async_client.create(input=text, **params)

        if not isinstance(response, dict):
            response = response.model_dump()
        return response["data"][0]["embedding"]



"""Utility functions for file operations."""

from __future__ import annotations

import base64
import os

from langchain_upstage.utils.constants import MEGABYTE


def file_to_base64_message(
    file_path: str,
    supported_extensions: list[str],
    max_file_size_bytes: int,
    allow_urls: bool = True,
) -> dict:
    """Convert file to base64 encoded message format.

    This is a generic utility function that can be used across different modules
    for converting files to the format expected by Upstage APIs.

    Args:
        file_path: Path to the file or URL to convert.
        supported_extensions: List of supported file extensions (e.g., ["pdf", "jpg"]).
        max_file_size_bytes: Maximum file size in bytes.
        allow_urls: Whether to allow URLs. If True, URLs are returned as-is without
            validation. If False, URLs will raise ValueError.

    Returns:
        dict: Dictionary with 'type' and 'image_url' keys. For URLs, returns the URL
            directly. For local files, returns base64 encoded data.

    Raises:
        FileNotFoundError: If the file does not exist (URLs are skipped if
            allow_urls=True).
        ValueError: If file format is unsupported, file is too large, or URL is
            not allowed when allow_urls=False.
    """

    if file_path.startswith(("http://", "https://")):
        if not allow_urls:
            raise ValueError("URLs are not allowed for this operation")
        return {"type": "image_url", "image_url": {"url": file_path}}

    if not os.path.exists(file_path):
        raise FileNotFoundError(f"File not found: {file_path}")

    ext = file_path.lower().split(".")[-1]
    if ext not in supported_extensions:
        supported = ", ".join([f".{e}" for e in supported_extensions])
        raise ValueError(
            f"Unsupported file format: .{ext}. Supported formats: {supported}"
        )

    file_size = os.path.getsize(file_path)
    if file_size > max_file_size_bytes:
        max_size_mb = max_file_size_bytes / MEGABYTE
        raise ValueError(
            f"File too large: {file_size / MEGABYTE:.2f}MB. "
            f"Maximum size: {max_size_mb}MB"
        )

    with open(file_path, "rb") as f:
        file_bytes = f.read()
        base64_data = base64.b64encode(file_bytes).decode("utf-8")

    return {
        "type": "image_url",
        "image_url": {"url": f"data:application/octet-stream;base64,{base64_data}"},
    }



from __future__ import annotations

import os
from typing import (
    Any,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Union,
)

import openai
from langchain_core.callbacks import (
    AsyncCallbackManagerForLLMRun,
    CallbackManagerForLLMRun,
)
from langchain_core.language_models.chat_models import LangSmithParams
from langchain_core.messages import AIMessageChunk, BaseMessage, HumanMessage
from langchain_core.messages.utils import convert_to_openai_messages
from langchain_core.outputs import ChatGenerationChunk, ChatResult
from langchain_core.utils import from_env, secret_from_env
from langchain_openai.chat_models.base import BaseChatOpenAI
from pydantic import Field, SecretStr, model_validator
from tokenizers import Tokenizer
from typing_extensions import Self

from langchain_upstage.document_parse import UpstageDocumentParseLoader

DOC_PARSING_MODEL = ["solar-pro2", "solar-pro3"]
SOLAR_TOKENIZERS = {
    "solar-pro3": "upstage/solar-pro3-tokenizer",
    "solar-pro2": "upstage/solar-pro2-tokenizer",
    "solar-mini": "upstage/solar-1-mini-tokenizer",
}

# Constants for header management
UPSTAGE_CLIENT_HEADER = "x-upstage-client"
UPSTAGE_CLIENT_VALUE = "langchain"
DEFAULT_HEADERS = {UPSTAGE_CLIENT_HEADER: UPSTAGE_CLIENT_VALUE}


class ChatUpstage(BaseChatOpenAI):
    """ChatUpstage chat model.

    To use, you should have the environment variable `UPSTAGE_API_KEY` set with your
    API key or pass it as a named parameter to the constructor.

    Example:
        ```python
        from langchain_upstage import ChatUpstage

        model = ChatUpstage()
        ```
    """

    @property
    def lc_secrets(self) -> Dict[str, str]:
        return {"upstage_api_key": "UPSTAGE_API_KEY"}

    @classmethod
    def get_lc_namespace(cls) -> List[str]:
        return ["langchain", "chat_models", "upstage"]

    @property
    def lc_attributes(self) -> Dict[str, Any]:
        attributes: Dict[str, Any] = {}

        if self.upstage_api_base:
            attributes["upstage_api_base"] = self.upstage_api_base

        return attributes

    @property
    def _llm_type(self) -> str:
        """Return type of chat model."""
        return "upstage-chat"

    def _get_ls_params(
        self, stop: Optional[List[str]] = None, **kwargs: Any
    ) -> LangSmithParams:
        """Get the parameters used to invoke the model."""
        params = super()._get_ls_params(stop=stop, **kwargs)
        params["ls_provider"] = "upstage"
        return params

    model_name: str = Field(default="solar-mini", alias="model")
    """Model name to use."""

    upstage_api_key: SecretStr = Field(
        default_factory=secret_from_env(
            "UPSTAGE_API_KEY",
            error_message=(
                "You must specify an api key. "
                "You can pass it an argument as `api_key=...` or "
                "set the environment variable `UPSTAGE_API_KEY`."
            ),
        ),
        alias="api_key",
    )
    """Automatically inferred from env are `UPSTAGE_API_KEY` if not provided."""

    upstage_api_base: Optional[str] = Field(
        default_factory=from_env(
            "UPSTAGE_API_BASE", default="https://api.upstage.ai/v1/solar"
        ),
        alias="base_url",
    )
    """Base URL path for API requests.
    
    Leave blank if not using a proxy or service emulator.
    """

    openai_api_key: Optional[SecretStr] = Field(default=None)
    """OpenAI api key is not supported for upstage. use `upstage_api_key` instead."""

    openai_api_base: Optional[str] = Field(default=None)
    """OpenAI api base is not supported for upstage. use `upstage_api_base` instead."""

    openai_organization: Optional[str] = Field(default=None)
    """OpenAI organization is not supported for upstage."""

    tiktoken_model_name: Optional[str] = None
    """Tiktoken is not supported for upstage."""

    tokenizer_name: Optional[str] = "upstage/solar-pro2-tokenizer"
    """Huggingface tokenizer name.
    
    Solar tokenizer is opened in huggingface https://huggingface.co/upstage/solar-pro-tokenizer
    """

    top_k: Optional[int] = Field(default=None)
    """Limit sampling to the top K most likely tokens. Lower values are more focused."""

    prompt_cache_key: Optional[str] = Field(default=None)
    """Unique key for prompt caching. Use distinct keys per conversational context."""

    default_headers: Union[Mapping[str, str], None] = DEFAULT_HEADERS
    """Add trace headers."""

    disabled_params: dict[str, Any] = Field(
        default_factory=lambda: {"parallel_tool_calls": None}
    )

    @model_validator(mode="after")
    def validate_environment(self) -> Self:
        """Validate that api key and python package exists in environment."""
        if self.n is not None and self.n < 1:
            raise ValueError("n must be at least 1.")
        if self.n is not None and self.n > 1 and self.streaming:
            raise ValueError("n must be 1 when streaming.")

        # Ensure x-upstage-client header is always set to "langchain"
        if self.default_headers is None:
            self.default_headers = DEFAULT_HEADERS
        else:
            # Create a copy to avoid modifying the original
            headers = dict(self.default_headers)
            headers[UPSTAGE_CLIENT_HEADER] = UPSTAGE_CLIENT_VALUE
            self.default_headers = headers

        client_params: dict = {
            "api_key": (
                self.upstage_api_key.get_secret_value()
                if self.upstage_api_key
                else None
            ),
            "base_url": self.upstage_api_base,
            "timeout": self.request_timeout,
            "default_headers": self.default_headers,
            "default_query": self.default_query,
        }
        if self.max_retries is not None:
            client_params["max_retries"] = self.max_retries

        if not self.root_client:
            sync_specific: dict = {"http_client": self.http_client}
            self.root_client = openai.OpenAI(**client_params, **sync_specific)

        if not self.client:
            self.client = self.root_client.chat.completions

        if not self.root_async_client:
            async_specific: dict = {"http_client": self.http_async_client}
            self.root_async_client = openai.AsyncOpenAI(
                **client_params, **async_specific
            )

        if not self.async_client:
            self.async_client = self.root_async_client.chat.completions

        return self

    def _get_tokenizer(self) -> Tokenizer:
        self.tokenizer_name = SOLAR_TOKENIZERS.get(self.model_name, self.tokenizer_name)
        return Tokenizer.from_pretrained(self.tokenizer_name)

    def get_token_ids(self, text: str) -> List[int]:
        """Get the tokens present in the text."""
        tokenizer = self._get_tokenizer()
        encode = tokenizer.encode(text, add_special_tokens=False)
        return encode.ids

    def get_num_tokens_from_messages(
        self, messages: Sequence[BaseMessage], tools: Sequence[Any] | None = None
    ) -> int:
        """Calculate num tokens for solar model."""
        tokenizer = self._get_tokenizer()
        tokens_per_message = 5  # <|im_start|>{role}\n{message}<|im_end|>
        tokens_prefix = 1  # <|startoftext|>
        tokens_suffix = 3  # <|im_start|>assistant\n

        num_tokens = 0

        num_tokens += tokens_prefix

        messages_dict = convert_to_openai_messages(messages)
        for message in messages_dict:
            num_tokens += tokens_per_message
            for key, value in message.items():
                # Cast str(value) in case the message value is not a string
                # This occurs with function messages
                num_tokens += len(
                    tokenizer.encode(str(value), add_special_tokens=False)
                )
        # every reply is primed with <|im_start|>assistant
        num_tokens += tokens_suffix
        return num_tokens

    def _create_chat_result(
        self,
        response: Union[dict, openai.BaseModel],
        generation_info: Optional[dict] = None,
    ) -> ChatResult:
        rtn = super()._create_chat_result(response, generation_info)

        if not isinstance(response, openai.BaseModel):
            return rtn

        choices = getattr(response, "choices", None)
        if choices and hasattr(choices[0].message, "reasoning_content"):
            reasoning_content = choices[0].message.reasoning_content
            if reasoning_content is not None:
                rtn.generations[0].message.additional_kwargs["reasoning_content"] = (
                    reasoning_content
                )
        elif choices and hasattr(choices[0].message, "model_extra"):
            model_extra = choices[0].message.model_extra
            if isinstance(model_extra, dict) and (
                reasoning := model_extra.get("reasoning")
            ):
                rtn.generations[0].message.additional_kwargs["reasoning_content"] = (
                    reasoning
                )

        return rtn

    def _convert_chunk_to_generation_chunk(
        self,
        chunk: dict,
        default_chunk_class: type,
        base_generation_info: Optional[dict],
    ) -> Optional[ChatGenerationChunk]:
        generation_chunk = super()._convert_chunk_to_generation_chunk(
            chunk,
            default_chunk_class,
            base_generation_info,
        )
        if (choices := chunk.get("choices")) and generation_chunk:
            top = choices[0]
            if isinstance(generation_chunk.message, AIMessageChunk):
                if (
                    reasoning_content := top.get("delta", {}).get("reasoning_content")
                ) is not None:
                    generation_chunk.message.additional_kwargs["reasoning_content"] = (
                        reasoning_content
                    )
                elif (reasoning := top.get("delta", {}).get("reasoning")) is not None:
                    generation_chunk.message.additional_kwargs["reasoning_content"] = (
                        reasoning
                    )

        return generation_chunk

    def _generate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        if self._using_doc_parsing_model(kwargs):
            document_contents = self._parse_documents(kwargs.pop("file_path"))
            messages.append(HumanMessage(document_contents))

        return super()._generate(messages, stop=stop, run_manager=run_manager, **kwargs)

    async def _agenerate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        if self._using_doc_parsing_model(kwargs):
            document_contents = self._parse_documents(kwargs.pop("file_path"))
            messages.append(HumanMessage(document_contents))

        return await super()._agenerate(
            messages, stop=stop, run_manager=run_manager, **kwargs
        )

    def _using_doc_parsing_model(self, kwargs: Dict[str, Any]) -> bool:
        if "file_path" in kwargs:
            if self.model_name in DOC_PARSING_MODEL:
                return True
            raise ValueError("file_path is not supported for this model.")
        return False

    def _parse_documents(self, file_path: str) -> str:
        document_contents = "Documents:\n"

        loader = UpstageDocumentParseLoader(
            api_key=(
                self.upstage_api_key.get_secret_value()
                if self.upstage_api_key
                else None
            ),
            file_path=file_path,
            output_format="text",
            coordinates=False,
        )
        docs = loader.load()

        if isinstance(file_path, list):
            file_titles = [os.path.basename(path) for path in file_path]
        else:
            file_titles = [os.path.basename(file_path)]

        for i, doc in enumerate(docs):
            file_title = file_titles[min(i, len(file_titles) - 1)]
            document_contents += f"{file_title}:\n{doc.page_content}\n\n"
        return document_contents



from __future__ import annotations

import json
import os
from typing import Any, Optional, Type, cast

from langchain_core.callbacks import (
    AsyncCallbackManagerForToolRun,
    CallbackManagerForToolRun,
)
from langchain_core.messages import HumanMessage
from langchain_core.tools import BaseTool
from langchain_core.utils import convert_to_secret_str
from openai import NOT_GIVEN
from pydantic import BaseModel, Field, SecretStr

from langchain_upstage import ChatUpstage
from langchain_upstage.utils.constants import MEGABYTE
from langchain_upstage.utils.file_utils import file_to_base64_message

INFORMATION_EXTRACT_BASE_URL = "https://api.upstage.ai/v1/information-extraction"
SUPPORTED_EXTENSIONS = [
    "jpeg",
    "png",
    "bmp",
    "pdf",
    "tiff",
    "heic",
    "docx",
    "pptx",
    "xlsx",
]
MAX_FILE_SIZE_MB = 50
MAX_FILE_SIZE_BYTES = MAX_FILE_SIZE_MB * MEGABYTE
DEFAULT_PAGES_PER_CHUNK = 5
DEFAULT_CONFIDENCE = True
DEFAULT_DOC_SPLIT = False
DEFAULT_LOCATION = False


class UpstageUniversalInformationExtractionInput(BaseModel):
    """Input for the Universal Information Extraction tool."""

    image_urls: list[str] = Field(
        description=(
            "List of file paths or URLs to images/documents to extract information "
            f"from. Supported formats: {', '.join(SUPPORTED_EXTENSIONS)}. "
            f"Maximum file size: {MAX_FILE_SIZE_MB}MB per file. "
            "URLs starting with http:// or https:// are supported."
        )
    )
    response_format: dict = Field(
        description="JSON schema defining the structure of information to extract. "
        "This should be a dict with 'type' and 'json_schema' keys, where 'json_schema' "
        "defines the expected output structure."
    )
    pages_per_chunk: int = Field(
        default=DEFAULT_PAGES_PER_CHUNK,
        description=(
            f"Number of pages to process per chunk. "
            f"Default is {DEFAULT_PAGES_PER_CHUNK}."
        ),
    )
    confidence: bool = Field(
        default=DEFAULT_CONFIDENCE,
        description=(
            f"Whether to include confidence scores in the response. "
            f"Default is {DEFAULT_CONFIDENCE}."
        ),
    )
    doc_split: bool = Field(
        default=DEFAULT_DOC_SPLIT,
        description=f"Whether to split documents. Default is {DEFAULT_DOC_SPLIT}.",
    )
    location: bool = Field(
        default=DEFAULT_LOCATION,
        description=(
            f"Whether to include location information. Default is {DEFAULT_LOCATION}."
        ),
    )


def _file_to_base64(file_path: str) -> dict:
    """Convert file to base64 encoded message format.

    Args:
        file_path: Path to the file or URL to convert.

    Returns:
        dict: Dictionary with 'type' and 'image_url' keys. For URLs, returns the
            URL directly. For local files, returns base64 encoded data.

    Raises:
        FileNotFoundError: If the file does not exist (URLs are skipped).
        ValueError: If the file format is unsupported or file size exceeds limit.
    """
    return file_to_base64_message(
        file_path,
        supported_extensions=SUPPORTED_EXTENSIONS,
        max_file_size_bytes=MAX_FILE_SIZE_BYTES,
        allow_urls=True,  # Universal extraction supports URLs
    )


class UpstageUniversalInformationExtraction(BaseTool):
    """Tool for extracting structured information from any document using universal
    extraction.

    This tool uses Upstage's universal information extraction model that can extract
    information from any document type based on a provided JSON schema.

    Unlike prebuilt models, this tool requires you to define a JSON schema that
    specifies what information to extract from the document. This makes it flexible
    for any document type but requires more configuration.

    To use, you should have the environment variable `UPSTAGE_API_KEY`
    set with your API key or pass it as a named parameter to the constructor.

    Args:
        model: The model to use for extraction. Defaults to "information-extract".
        api_key: Optional API key. If not provided, will use UPSTAGE_API_KEY
            environment variable.

    Example:
        .. code-block:: python

            from langchain_upstage import UpstageUniversalInformationExtractionTool

            tool = UpstageUniversalInformationExtractionTool()

            # Use with a schema
            result = tool.invoke({
                "image_urls": ["/path/to/document.pdf"],
                "response_format": {
                    "type": "json_schema",
                    "json_schema": {
                        "name": "extraction_schema",
                        "schema": {
                            "type": "object",
                            "properties": {
                                "field_name": {"type": "string"}
                            }
                        }
                    }
                }
            })
    """

    name: str = "universal_information_extraction"
    description: str = (
        "Extract structured information from any document using a universal "
        "extraction model. This tool can extract information from any document "
        "type by providing a JSON schema that defines what information to extract. "
        "Use this tool when you need to extract custom information from documents "
        "that don't match prebuilt models, or when you need flexible extraction "
        "based on a schema you define."
    )

    model: str = Field(default="information-extract")
    upstage_api_key: Optional[SecretStr] = Field(default=None, alias="api_key")
    api_wrapper: Optional[ChatUpstage] = None

    args_schema: Type[BaseModel] = UpstageUniversalInformationExtractionInput

    def __init__(
        self,
        model: str = "information-extract",
        api_key: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        upstage_api_key = kwargs.get("upstage_api_key", None)
        if not upstage_api_key:
            upstage_api_key = api_key
        if not upstage_api_key:
            upstage_api_key = SecretStr(os.getenv("UPSTAGE_API_KEY", ""))
        upstage_api_key = convert_to_secret_str(upstage_api_key)

        if (
            not upstage_api_key
            or not upstage_api_key.get_secret_value()
            or upstage_api_key.get_secret_value() == ""
        ):
            raise ValueError("UPSTAGE_API_KEY must be set or passed")

        # Universal IE uses OpenAI Chat Completion API format
        # So we can use ChatUpstage with custom base_url
        api_wrapper = ChatUpstage(
            model=model,
            api_key=upstage_api_key,
            base_url=INFORMATION_EXTRACT_BASE_URL,
        )

        super().__init__(
            model=model,
            upstage_api_key=upstage_api_key,
            api_wrapper=api_wrapper,
            **kwargs,
        )

    def _run(
        self,
        image_urls: list[str],
        response_format: dict,
        pages_per_chunk: int = DEFAULT_PAGES_PER_CHUNK,
        confidence: bool = DEFAULT_CONFIDENCE,
        doc_split: bool = DEFAULT_DOC_SPLIT,
        location: bool = DEFAULT_LOCATION,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> dict:
        """Extract information from documents using the universal extraction model.

        Args:
            image_urls: List of file paths or URLs to images/documents to extract
                from.
            response_format: JSON schema defining the structure of information to
                extract. Should be a dict with 'type' and 'json_schema' keys.
            pages_per_chunk: Number of pages to process per chunk. Default is 5.
            confidence: Whether to include confidence scores in the response.
                Default is True.
            doc_split: Whether to split documents. Default is False.
            location: Whether to include location information. Default is False.
            run_manager: Optional callback manager for tool execution.

        Returns:
            dict: Extracted information as a dictionary. If the response is JSON,
                returns the parsed JSON. Otherwise, returns a dict with 'content' key.

        Raises:
            ValueError: If API wrapper is not initialized.
            FileNotFoundError: If any file does not exist.
            ValueError: If file format is unsupported or file is too large.
            Exception: If API call fails or response parsing fails.
        """
        if self.api_wrapper is None:
            error_msg = (
                "API wrapper not initialized. Tool may not have been properly "
                "configured."
            )
            if run_manager:
                run_manager.on_tool_error(ValueError(error_msg))
            raise ValueError(error_msg)

        try:
            # Convert files to base64
            contents: list[dict[str, Any]] = [
                _file_to_base64(url) for url in image_urls
            ]

            # Prepare messages
            # Type cast: contents is list[dict] which is compatible with
            # list[str | dict[Any, Any]]
            messages = [HumanMessage(content=cast(Any, contents))]

            # Use ChatUpstage to make the API call
            api_wrapper = cast(ChatUpstage, self.api_wrapper)
            response = api_wrapper.invoke(
                messages,
                response_format=response_format,
                stream=NOT_GIVEN,
                extra_body={
                    "chunking": {"pages_per_chunk": pages_per_chunk},
                    "confidence": confidence,
                    "doc_split": doc_split,
                    "location": location,
                },
            )

            # Parse response
            if hasattr(response, "content") and response.content:
                # Type check: response.content should be str for json.loads
                content_str = (
                    response.content
                    if isinstance(response.content, str)
                    else str(response.content)
                )
                try:
                    return json.loads(content_str)
                except json.JSONDecodeError:
                    return {"content": content_str}
            return {}
        except (FileNotFoundError, ValueError) as e:
            # Re-raise validation errors as-is
            if run_manager:
                run_manager.on_tool_error(e)
            raise
        except Exception as e:
            # Wrap other exceptions with context
            error_msg = f"Failed to extract information from documents: {str(e)}"
            if run_manager:
                run_manager.on_tool_error(Exception(error_msg))
            raise Exception(error_msg) from e

    async def _arun(
        self,
        image_urls: list[str],
        response_format: dict,
        pages_per_chunk: int = DEFAULT_PAGES_PER_CHUNK,
        confidence: bool = DEFAULT_CONFIDENCE,
        doc_split: bool = DEFAULT_DOC_SPLIT,
        location: bool = DEFAULT_LOCATION,
        run_manager: Optional[AsyncCallbackManagerForToolRun] = None,
    ) -> dict:
        """Extract information from documents asynchronously using the universal
        extraction model.

        Args:
            image_urls: List of file paths or URLs to images/documents to extract
                from.
            response_format: JSON schema defining the structure of information to
                extract. Should be a dict with 'type' and 'json_schema' keys.
            pages_per_chunk: Number of pages to process per chunk. Default is 5.
            confidence: Whether to include confidence scores in the response.
                Default is True.
            doc_split: Whether to split documents. Default is False.
            location: Whether to include location information. Default is False.
            run_manager: Optional async callback manager for tool execution.

        Returns:
            dict: Extracted information as a dictionary. If the response is JSON,
                returns the parsed JSON. Otherwise, returns a dict with 'content' key.

        Raises:
            ValueError: If API wrapper is not initialized.
            FileNotFoundError: If any file does not exist.
            ValueError: If file format is unsupported or file is too large.
            Exception: If API call fails or response parsing fails.
        """
        if self.api_wrapper is None:
            error_msg = (
                "API wrapper not initialized. Tool may not have been properly "
                "configured."
            )
            if run_manager:
                await run_manager.on_tool_error(ValueError(error_msg))
            raise ValueError(error_msg)

        try:
            # Convert files to base64
            contents: list[dict[str, Any]] = [
                _file_to_base64(url) for url in image_urls
            ]

            # Prepare messages
            # Type cast: contents is list[dict] which is compatible with
            # list[str | dict[Any, Any]]
            messages = [HumanMessage(content=cast(Any, contents))]

            # Use ChatUpstage to make the API call
            api_wrapper = cast(ChatUpstage, self.api_wrapper)
            response = await api_wrapper.ainvoke(
                messages,
                response_format=response_format,
                extra_body={
                    "chunking": {"pages_per_chunk": pages_per_chunk},
                    "confidence": confidence,
                    "doc_split": doc_split,
                    "location": location,
                },
            )

            # Parse response
            if hasattr(response, "content") and response.content:
                # Type check: response.content should be str for json.loads
                content_str = (
                    response.content
                    if isinstance(response.content, str)
                    else str(response.content)
                )
                try:
                    return json.loads(content_str)
                except json.JSONDecodeError:
                    return {"content": content_str}
            return {}
        except (FileNotFoundError, ValueError) as e:
            # Re-raise validation errors as-is
            if run_manager:
                await run_manager.on_tool_error(e)
            raise
        except Exception as e:
            # Wrap other exceptions with context
            error_msg = f"Failed to extract information from documents: {str(e)}"
            if run_manager:
                await run_manager.on_tool_error(Exception(error_msg))
            raise Exception(error_msg) from e

    def generate_schema(self, image_urls: list[str]) -> dict:
        """Generate JSON schema from documents.

        This method analyzes the provided documents and generates a JSON schema
        that can be used for information extraction.

        Args:
            image_urls: List of file paths or URLs to images/documents to analyze.
                Maximum 3 images are supported.

        Returns:
            dict: Generated JSON schema in the format:
                {
                    "type": "json_schema",
                    "json_schema": {
                        "name": "schema_name",
                        "schema": {...}
                    }
                }

        Raises:
            ValueError: If API wrapper is not initialized or too many images provided.
            FileNotFoundError: If any file does not exist.
            ValueError: If file format is unsupported or file is too large.
            Exception: If API call fails or response parsing fails.
        """
        if self.api_wrapper is None:
            raise ValueError(
                "API wrapper not initialized. Tool may not have been properly "
                "configured."
            )

        if len(image_urls) > 3:
            raise ValueError("Maximum 3 images are supported for schema generation.")

        try:
            # Convert files to base64
            contents: list[dict[str, Any]] = [
                _file_to_base64(url) for url in image_urls
            ]

            # Prepare messages
            # Type cast: contents is list[dict] which is compatible with
            # list[str | dict[Any, Any]]
            messages = [HumanMessage(content=cast(Any, contents))]

            # Create ChatUpstage instance for schema generation endpoint
            # The endpoint is /schema-generation/chat/completions
            # Get API key from api_wrapper if upstage_api_key is not set
            api_key = self.upstage_api_key
            if not api_key and self.api_wrapper:
                api_key = getattr(self.api_wrapper, "upstage_api_key", None)

            if not api_key:
                raise ValueError("API key not available for schema generation")

            schema_api_wrapper = ChatUpstage(
                model=self.model,
                api_key=api_key,
                base_url=f"{INFORMATION_EXTRACT_BASE_URL}/schema-generation",
            )

            # Make API call
            response = schema_api_wrapper.invoke(messages)

            # Parse response - content contains stringified JSON schema
            if hasattr(response, "content") and response.content:
                # Type check: response.content should be str for json.loads
                content_str = (
                    response.content
                    if isinstance(response.content, str)
                    else str(response.content)
                )
                try:
                    # Parse the JSON schema from content
                    schema_content = json.loads(content_str)
                    # Return in the format expected by extract() method
                    return {"type": "json_schema", "json_schema": schema_content}
                except json.JSONDecodeError:
                    # If content is not valid JSON, try to wrap it
                    return {
                        "type": "json_schema",
                        "json_schema": {
                            "name": "generated_schema",
                            "schema": json.loads(content_str),
                        },
                    }
            raise ValueError("Empty response from schema generation API")

        except (FileNotFoundError, ValueError):
            raise
        except Exception as e:
            raise Exception(f"Failed to generate schema: {str(e)}") from e

    async def agenerate_schema(self, image_urls: list[str]) -> dict:
        """Generate JSON schema from documents asynchronously.

        Args:
            image_urls: List of file paths or URLs to images/documents to analyze.
                Maximum 3 images are supported.

        Returns:
            dict: Generated JSON schema.

        Raises:
            ValueError: If API wrapper is not initialized or too many images provided.
            FileNotFoundError: If any file does not exist.
            ValueError: If file format is unsupported or file is too large.
            Exception: If API call fails or response parsing fails.
        """
        if self.api_wrapper is None:
            raise ValueError(
                "API wrapper not initialized. Tool may not have been properly "
                "configured."
            )

        if len(image_urls) > 3:
            raise ValueError("Maximum 3 images are supported for schema generation.")

        try:
            # Convert files to base64
            contents: list[dict[str, Any]] = [
                _file_to_base64(url) for url in image_urls
            ]

            # Prepare messages
            # Type cast: contents is list[dict] which is compatible with
            # list[str | dict[Any, Any]]
            messages = [HumanMessage(content=cast(Any, contents))]

            # Create ChatUpstage instance for schema generation endpoint
            # Get API key from api_wrapper if upstage_api_key is not set
            api_key = self.upstage_api_key
            if not api_key and self.api_wrapper:
                api_key = getattr(self.api_wrapper, "upstage_api_key", None)

            if not api_key:
                raise ValueError("API key not available for schema generation")

            schema_api_wrapper = ChatUpstage(
                model=self.model,
                api_key=api_key,
                base_url=f"{INFORMATION_EXTRACT_BASE_URL}/schema-generation",
            )

            # Make async API call
            response = await schema_api_wrapper.ainvoke(messages)

            # Parse response
            if hasattr(response, "content") and response.content:
                # Type check: response.content should be str for json.loads
                content_str = (
                    response.content
                    if isinstance(response.content, str)
                    else str(response.content)
                )
                try:
                    schema_content = json.loads(content_str)
                    return {"type": "json_schema", "json_schema": schema_content}
                except json.JSONDecodeError:
                    return {
                        "type": "json_schema",
                        "json_schema": {
                            "name": "generated_schema",
                            "schema": json.loads(content_str),
                        },
                    }
            raise ValueError("Empty response from schema generation API")

        except (FileNotFoundError, ValueError):
            raise
        except Exception as e:
            raise Exception(f"Failed to generate schema: {str(e)}") from e



import logging
import os
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Union

from langchain_core.document_loaders import BaseLoader, Blob
from langchain_core.documents import Document

from langchain_upstage.document_parse_parsers import (
    DOCUMENT_PARSE_BASE_URL,
    DOCUMENT_PARSE_DEFAULT_MODEL,
    OCR,
    Category,
    OutputFormat,
    SplitType,
    UpstageDocumentParseParser,
)
from langchain_upstage.utils.value_retriever import get_from_param_or_env

logger = logging.getLogger("pypdf")
logger.setLevel(logging.ERROR)


def validate_file_path(file_path: Union[str, Path, List[str], List[Path]]) -> None:
    """Validates if a file exists at the given file path.

    Args:
        file_path: The file path(s) to be validated.

    Raises:
        FileNotFoundError: If the file or any of the files in the list do not exist.
    """
    if isinstance(file_path, list):
        for path in file_path:
            validate_file_path(path)
        return
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"File not found: {file_path}")


class UpstageDocumentParseLoader(BaseLoader):
    """Upstage Document Parse Loader.

    To use, you should have the environment variable `UPSTAGE_API_KEY`
    set with your API key or pass it as a named parameter to the constructor.

    Example:
        ```python
        from langchain_upstage import UpstageDocumentParseLoader

        file_path = "/PATH/TO/YOUR/FILE.pdf"
        loader = UpstageDocumentParseLoader(
            file_path, split="page", output_format="text"
        )
        ```
    """

    def __init__(
        self,
        file_path: Union[str, Path, List[str], List[Path]],
        split: SplitType = "none",
        api_key: Optional[str] = None,
        base_url: str = DOCUMENT_PARSE_BASE_URL,
        model: str = DOCUMENT_PARSE_DEFAULT_MODEL,
        chart_recognition: bool = True,
        ocr: OCR = "auto",
        output_format: OutputFormat = "html",
        coordinates: bool = True,
        base64_encoding: Optional[List[Category]] = None,
    ):
        """
        Initializes an instance of the Upstage document parse loader.

        Args:
            file_path: The path to the document to be loaded.
            split (SplitType, optional): The type of splitting to be applied.
            api_key (str, optional): The API key for accessing the Upstage API.
                Defaults to None, in which case it will be fetched from the environment
                variable `UPSTAGE_API_KEY`.
            base_url (str, optional): The base URL for accessing the Upstage API.
            model (str): The model to be used for the document parse.
                Defaults to `'document-parse'`.
            ocr (OCRMode, optional): Extract text from images in the document using
                OCR.

                If the value is `'force'`, OCR is used to extract text from an image.

                If the value is `'auto'`, text is extracted from a PDF. (An error will
                occur if the value is `'auto'` and the input is NOT in PDF format)
            output_format (OutputFormat, optional): Format of the inference results.
            coordinates (bool, optional): Whether to include the coordinates of the
                OCR in the output.
            base64_encoding (List[Category], optional): The category of the elements to
                be encoded in base64.
        """
        self.file_path = file_path
        self.split = split
        self.api_key = get_from_param_or_env(
            "UPSTAGE_API_KEY",
            api_key,
            "UPSTAGE_API_KEY",
            os.environ.get("UPSTAGE_API_KEY"),
        )
        self.base_url = base_url
        self.model = model
        self.chart_recognition = chart_recognition
        self.ocr = ocr
        self.output_format = output_format
        self.coordinates = coordinates
        self.base64_encoding = base64_encoding if base64_encoding is not None else []
        self.parser = UpstageDocumentParseParser(
            api_key=self.api_key,
            base_url=self.base_url,
            model=self.model,
            split=self.split,
            ocr=self.ocr,
            output_format=self.output_format,
            coordinates=self.coordinates,
            base64_encoding=self.base64_encoding,
        )

        validate_file_path(self.file_path)

    def load(self) -> List[Document]:
        """Loads and parses the document using the `UpstageDocumentParseParser`.

        Returns:
            `Document` objects representing the parsed layout analysis.
        """

        if isinstance(self.file_path, list):
            result = []

            for file_path in self.file_path:
                blob = Blob.from_path(file_path)
                result.extend(list(self.parser.lazy_parse(blob, is_batch=True)))

            return result

        else:
            blob = Blob.from_path(self.file_path)
            return list(self.parser.lazy_parse(blob, is_batch=True))

    def lazy_load(self) -> Iterator[Document]:
        """Lazily loads and parses the document using the `UpstageDocumentParseParser`.

        Returns:
            An iterator of `Document` objects representing the parsed layout analysis.
        """

        if isinstance(self.file_path, list):
            for file_path in self.file_path:
                blob = Blob.from_path(file_path)
                yield from self.parser.lazy_parse(blob, is_batch=True)
        else:
            blob = Blob.from_path(self.file_path)

            yield from self.parser.lazy_parse(blob)

    def merge_and_split(
        self, documents: List[Document], splitter: Optional[object] = None
    ) -> List[Document]:
        """
        Merges the page content and metadata of multiple documents into a single
        document, or splits the documents using a custom splitter.

        Args:
            documents (list): A list of `Document` objects to be merged and split.
            splitter (object, optional): An optional splitter object that implements the
                `split_documents` method. If provided, the documents will be split using
                this splitter. Defaults to None, in which case the documents are merged.

        Returns:
            list: A list of `Document` objects. If no splitter is provided, a single
            `Document` object is returned with the merged content and combined metadata.
            If a splitter is provided, the documents are split and a list of `Document`
            objects is returned.

        Raises:
            AssertionError: If a splitter is provided but it does not implement
                the `split_documents` method.
        """
        if splitter is None:
            merged_content = " ".join([doc.page_content for doc in documents])

            metadatas: Dict[str, Any] = dict()
            for _meta in [doc.metadata for doc in documents]:
                for key, value in _meta.items():
                    if key in metadatas:
                        metadatas[key].append(value)
                    else:
                        metadatas[key] = [value]

            return [Document(page_content=merged_content, metadata=metadatas)]
        else:
            assert hasattr(splitter, "split_documents"), (
                "splitter must implement split_documents method"
            )

            return splitter.split_documents(documents)
